@@ -2,7 +2,7 @@
  *
  * Simple mark&sweep garbage collector.
  *
- * $Id: gc.h,v 1.7 2001/03/25 12:09:45 nisse Exp $ */
+ * $Id: gc.h,v 1.11 2001/09/27 12:09:52 nisse Exp $ */
 
 /* lsh, an implementation of the ssh protocol
  *
@@ -31,13 +31,10 @@
 void gc_register(struct lsh_object *o);
 void gc_kill(struct lsh_object *o);
 
-void gc(struct lsh_object *root);
-void gc_maybe(struct lsh_object *root, int busy);
+void gc_global(struct resource *o);
+
+void gc(void);
 
-#if DEBUG_ALLOC
 void gc_final(void);
-#else
-#define gc_final()
-#endif
 
 #endif /* LSH_GC_H_INCLUDED */
