@@ -2,7 +2,7 @@
  *
  * client main program
  *
- * $Id: lsh.c,v 1.136 2001/03/25 12:21:17 nisse Exp $ */
+ * $Id: lsh.c,v 1.164 2002/06/27 17:18:32 nisse Exp $ */
 
 /* lsh, an implementation of the ssh protocol
  *
@@ -43,10 +43,10 @@
 #include "gateway_commands.h"
 #include "handshake.h"
 #include "lookup_verifier.h"
+#include "publickey_crypto.h"
 #include "randomness.h"
-#include "rsa.h"
 #include "sexp.h"
-#include "spki_commands.h"
+#include "spki.h"
 #include "srp.h" 
 #include "ssh.h"
 #include "tcpforward_commands.h"
@@ -68,17 +68,14 @@
 #include "lsh_argp.h"
 
 /* Forward declarations */
-struct command_simple lsh_verifier_command;
+struct command_2 lsh_verifier_command;
 #define OPTIONS2VERIFIER (&lsh_verifier_command.super.super)
 
-struct command_simple lsh_login_command;
+struct command_2 lsh_login_command;
 #define LSH_LOGIN (&lsh_login_command.super.super)
 
-static struct command options2known_hosts;
-#define OPTIONS2KNOWN_HOSTS (&options2known_hosts.super)
-
-struct command_simple options2service;
-#define OPTIONS2SERVICE (&options2service.super.super)
+struct command options2service;
+#define OPTIONS2SERVICE (&options2service.super)
 
 static struct command options2identities;
 #define OPTIONS2IDENTITIES (&options2identities.super)
@@ -87,9 +84,6 @@ static struct request_service request_userauth_service =
 STATIC_REQUEST_SERVICE(ATOM_SSH_USERAUTH);
 #define REQUEST_USERAUTH_SERVICE (&request_userauth_service.super.super)
 
-static struct request_service request_connection_service =
-STATIC_REQUEST_SERVICE(ATOM_SSH_CONNECTION);
-#define REQUEST_CONNECTION_SERVICE (&request_connection_service.super.super)
 
 #include "lsh.c.x"
 
@@ -108,8 +102,6 @@ STATIC_REQUEST_SERVICE(ATOM_SSH_CONNECTION);
      (vars
        (algorithms object algorithms_options)
 
-       (random object randomness_with_poll)
-
        (signature_algorithms object alist)
        (home . "const char *")
        
@@ -119,9 +111,6 @@ STATIC_REQUEST_SERVICE(ATOM_SSH_CONNECTION);
        (with_srp_keyexchange . int)
        (with_dh_keyexchange . int)
 
-       ; Ask for the userauth service
-       (with_userauth . int)
-
        ; Command to invoke to start ssh-connection service)
        (service object command)
        
@@ -139,21 +128,21 @@ STATIC_REQUEST_SERVICE(ATOM_SSH_CONNECTION);
 
 
 static struct lsh_options *
-make_options(struct io_backend *backend,
-	     struct exception_handler *handler,
+make_options(struct exception_handler *handler,
 	     int *exit_code)
 {
   NEW(lsh_options, self);
-  init_client_options(&self->super, backend, handler, exit_code);
+  const char *home = getenv("HOME");
+  struct randomness *r = make_user_random(home);
 
-  self->algorithms
-    = make_algorithms_options(all_symmetric_algorithms());
+  init_client_options(&self->super, 
+		      r,
+		      handler, exit_code);
 
-  self->random = make_default_random(NULL, handler);
-  
-  self->home = getenv("HOME");
-  
-  self->signature_algorithms = all_signature_algorithms(&self->random->super);
+  self->algorithms = make_algorithms_options(all_symmetric_algorithms());
+  self->home = home;
+
+  self->signature_algorithms = all_signature_algorithms(r); /* OK to init with NULL */
 
   self->sloppy = 0;
   self->capture = NULL;
@@ -169,8 +158,6 @@ make_options(struct io_backend *backend,
 
   /* By default, enable only one of dh and srp. */
   self->with_dh_keyexchange = -1;
-
-  self->with_userauth = -1;
   
   self->kex_algorithms = NULL;
   
@@ -180,125 +167,228 @@ make_options(struct io_backend *backend,
 
 /* Request ssh-userauth or ssh-connection service, as appropriate,
  * and pass the options as a first argument. */
-DEFINE_COMMAND_SIMPLE(options2service, a)
+DEFINE_COMMAND(options2service)
+     (struct command *s UNUSED,
+      struct lsh_object *a,
+      struct command_continuation *c,
+      struct exception_handler *e UNUSED)
 {
   CAST(lsh_options, options, a);
-  return &options->service->super;
+  COMMAND_RETURN(c, options->service);
 }
 
 /* Open hostkey database. By default, "~/.lsh/known_hosts". */
 
-static void
-do_options2known_hosts(struct command *ignored UNUSED,
-		       struct lsh_object *a,
-		       struct command_continuation *c,
-		       struct exception_handler *e)
+static struct spki_context *
+read_known_hosts(struct lsh_options *options)
 {
-  CAST(lsh_options, options, a);
-  
   struct lsh_string *tmp = NULL;
+  struct lsh_string *contents;
   const char *s = NULL;
-  struct lsh_fd *f;
+  int fd;
+  struct simple_buffer buffer;
+  struct spki_context *context;
+
+  context = make_spki_context(options->signature_algorithms);
   
   if (options->known_hosts)
     s = options->known_hosts;
-  else 
-    {
-      tmp = ssh_cformat("%lz/.lsh/known_hosts", options->home);
-      s = tmp->data;
-    }
-  
-  f = io_read_file(options->super.backend, s, e);
-
-  if (!f)
-    {
-      werror("Failed to open '%z' (errno = %i): %z.\n",
-	     s, errno, STRERROR(errno));
-      COMMAND_RETURN(c, make_spki_context(options->signature_algorithms));
-    }
   else
     {
-      CAST_SUBTYPE(command, read,
-		   make_spki_read_acls(options->signature_algorithms));
-      COMMAND_CALL(read, f, c, e);
+      tmp = ssh_format("%lz/.lsh/known_hosts", options->home);
+      s = lsh_get_cstring(tmp);
     }
+
+  fd = open(s, O_RDONLY);
+  if (fd < 0)
+    {
+      lsh_string_free(tmp);
+      werror("Failed to open `%z' for reading: %z\n",
+             s, STRERROR(errno));
+      return context;
+    }
+
   lsh_string_free(tmp);
+
+  contents = io_read_file_raw(fd, 5000);
+  
+  if (!contents)
+  {
+    werror("Failed to read known_hosts file: %z\n",
+           STRERROR(errno));
+    close(fd);
+    return context;
+  }
+
+  close(fd);
+
+  simple_buffer_init(&buffer, contents->length, contents->data);
+  while (!parse_eod(&buffer))
+    {
+      struct sexp *expr = sexp_parse(SEXP_TRANSPORT, &buffer);
+
+      if (!expr)
+        {
+          werror("read_known_hosts: S-expression syntax error.\n");
+          break;
+        }
+      if (!spki_add_acl(context, expr))
+        {
+          werror("read_known_hosts: Invalid ACL.\n");
+          break;
+        }
+    }
+  lsh_string_free(contents);
+  return context;
 }
 
-static struct command options2known_hosts =
-STATIC_COMMAND(do_options2known_hosts);
+/* For now, supports only a single key */
+static struct object_list *
+read_user_keys(struct lsh_options *options)
+{
+  struct lsh_string *tmp = NULL;
+  struct lsh_string *contents;
+  const char *name = NULL;
+  struct sexp *expr;
+  int fd;
+  int algorithm_name;
+  
+  struct signer *s;
+  struct verifier *v;
+  struct lsh_string *spki_public;
+
+  trace("read-user_keys\n");
+  
+  if (!options->with_publickey)
+    return make_object_list(0, -1);
+
+  if (options->identity)
+    name = options->identity;
+  else 
+    {
+      tmp = ssh_format("%lz/.lsh/identity", options->home);
+      name = lsh_get_cstring(tmp);
+    }
+
+  fd = open(name, O_RDONLY);
+  if (fd < 0)
+    {
+      werror("Failed to open `%z' for reading: %z\n",
+             name, STRERROR(errno));
+      lsh_string_free(tmp);
+
+      return make_object_list(0, -1);
+    }
+
+  lsh_string_free(tmp);
+
+  contents = io_read_file_raw(fd, 2000);
+
+  if (!contents)
+    {
+      werror("Failed to read private key file: %z\n",
+             STRERROR(errno));
+      close(fd);
+
+      return make_object_list(0, -1);
+    }
+
+  close(fd);
+
+  expr = string_to_sexp(SEXP_TRANSPORT, contents, 1);
+
+  if (!expr)
+    {
+      werror("S-expression syntax error in private key.\n");
+      return make_object_list(0, -1);
+    }
+
+  if (options->super.tty)
+    {
+      struct alist *mac = make_alist(0, -1);
+      struct alist *crypto = make_alist(0, -1);
+
+      alist_select_l(mac, options->algorithms->algorithms,
+		     2, ATOM_HMAC_SHA1, ATOM_HMAC_MD5, -1);
+      alist_select_l(crypto, options->algorithms->algorithms,
+		     4, ATOM_3DES_CBC, ATOM_BLOWFISH_CBC,
+		     ATOM_TWOFISH_CBC, ATOM_AES256_CBC, -1);
+      
+      expr = spki_pkcs5_decrypt(mac, crypto,
+                                options->super.tty,
+                                expr);
+      if (!expr)
+        {
+          werror("Decrypting private key failed.\n");
+          return make_object_list(0, -1);
+        }
+    }
+  
+  s = spki_sexp_to_signer(options->signature_algorithms, expr,
+                          &algorithm_name);
+  if (!s)
+    {
+      werror("Invalid private key.\n");
+      return make_object_list(0, -1);
+    }
+  
+  v = SIGNER_GET_VERIFIER(s);
+  assert(v);
+  
+  spki_public = sexp_format(spki_make_public_key(v),
+			    SEXP_CANONICAL, 0);
+
+  /* Test key here? */  
+  switch (algorithm_name)
+    {	  
+    case ATOM_DSA:
+      return make_object_list(2,
+                              make_keypair(ATOM_SSH_DSS,
+                                           PUBLIC_KEY(v),
+                                           s),
+                              make_keypair(ATOM_SPKI_SIGN_DSS,
+                                           spki_public, s),
+                              -1);
+      break;
+
+    case ATOM_RSA_PKCS1:
+    case ATOM_RSA_PKCS1_SHA1:
+      return make_object_list(2, 
+                              make_keypair(ATOM_SSH_RSA,
+                                           PUBLIC_KEY(v),
+                                           s),
+                              make_keypair(ATOM_SPKI_SIGN_RSA,
+                                           spki_public, s),
+                              -1);
+
+    case ATOM_RSA_PKCS1_MD5:
+      return make_object_list(1,
+                              make_keypair(ATOM_SPKI_SIGN_RSA,
+                                           spki_public, s),
+                              -1);
+
+    default:
+      fatal("Internal error!\n");
+    }
+}
+
+/* FIXME: Call read_user_key directly from main instead. */
 
 /* Read user's private keys. By default, "~/.lsh/identity". */
 static void
 do_options2identities(struct command *ignored UNUSED,
 		      struct lsh_object *a,
 		      struct command_continuation *c,
-		      struct exception_handler *e)
+		      struct exception_handler *e UNUSED)
 {
   CAST(lsh_options, options, a);
-  
-  struct lsh_string *tmp = NULL;
-  const char *s = NULL;
-  struct lsh_fd *f = NULL;
 
-  trace("do_options2identities\n");
-  
-  if (!options->with_publickey)
-    {
-      COMMAND_RETURN(c, make_object_list(0, -1));
-      return;
-    }
-  if (options->identity)
-    s = options->identity;
-  else 
-    {
-      tmp = ssh_cformat("%lz/.lsh/identity", options->home);
-      s = tmp->data;
-    }
-  
-  f = io_read_file(options->super.backend, s, e);
-  
-  if (!f)
-    {
-      werror("Failed to open '%z' (errno = %i): %z.\n",
-	     s, errno, STRERROR(errno));
-      COMMAND_RETURN(c, make_object_list(0, -1));
-    }
-  else
-    {
-      struct command *command
-	= make_spki_read_userkeys(options->algorithms->algorithms,
-				  options->signature_algorithms,
-				  options->super.tty);
-      COMMAND_CALL(command, f, c, e);
-    }
-  lsh_string_free(tmp);
+  COMMAND_RETURN(c, read_user_keys(options));
 }
 
 static struct command options2identities =
 STATIC_COMMAND(do_options2identities);
 
-/* GABA:
-   (class
-     (name options_command)
-     (super command)
-     (vars
-       (options object lsh_options)))
-*/
-
-static struct command *
-make_options_command(struct lsh_options *options,
-		     void (*call)(struct command *s,
-				  struct lsh_object *a,
-				  struct command_continuation *c,
-				  struct exception_handler *e))
-{
-  NEW(options_command, self);
-  self->super.call = call;
-  self->options = options;
-
-  return &self->super;
-}
 
 /* Maps a host key to a (trusted) verifier object. */
 
@@ -316,7 +406,7 @@ make_options_command(struct lsh_options *options,
        ; If non-null, append an ACL for the received key to this file.
        (file object abstract_write)
        ; For fingerprinting
-       (hash object hash_algorithm)))
+       (hash const object hash_algorithm)))
 */
 
 static struct verifier *
@@ -419,10 +509,12 @@ do_lsh_lookup(struct lookup_verifier *c,
     }
   else
     {
+      struct sexp *acl;
+      
       verbose("SPKI authorization failed.\n");
       if (!self->sloppy)
 	{
-	  werror("lsh: Server's hostkey is not trusted. Disconnecting.\n");
+	  werror("Server's hostkey is not trusted. Disconnecting.\n");
 	  return NULL;
 	}
       
@@ -430,34 +522,62 @@ do_lsh_lookup(struct lookup_verifier *c,
       if (!quiet_flag)
 	{
 	  /* Display fingerprint */
-	  struct lsh_string *fingerprint
-	    = hash_string(self->hash,
-			  sexp_format(subject->key, SEXP_CANONICAL, 0),
-			  1);
-			  
+
+	  struct lsh_string *spki_fingerprint = 
+	    hash_string(self->hash,
+			sexp_format(subject->key, SEXP_CANONICAL, 0),
+			1);
+
+
+	  struct lsh_string *fingerprint = 
+	    lsh_string_colonize( 
+				ssh_format( "%lfxS", 
+					    hash_string(&crypto_md5_algorithm,
+							key,
+							0)
+					    ), 
+				2, 
+				1  
+				);
+
+	  struct lsh_string *babble = 
+	    lsh_string_bubblebabble( 
+				    hash_string(&crypto_sha1_algorithm,
+						key,
+						0),
+				    1 
+				    );
+	  
 	  if (!INTERACT_YES_OR_NO
 	      (self->tty,
 	       ssh_format("Received unauthenticated key for host %lS\n"
-			  "Fingerprint: %lfxS\n"
+			  "Key details:\n"
+			  "Bubble Babble: %lfS\n"
+			  "Fingerprint:   %lfS\n"
+			  "SPKI SHA1:     %lfxS\n"
 			  "Do you trust this key? (y/n) ",
-			  self->host->ip, fingerprint), 0, 1))
+			  self->host->ip, babble, fingerprint, spki_fingerprint), 0, 1))
 	    return NULL;
 	}
+
+      acl = sexp_l(2, sexp_a(ATOM_ACL),
+		   sexp_l(3, sexp_a(ATOM_ENTRY),
+			  subject->key,
+			  sexp_l(2, sexp_a(ATOM_TAG),
+				 self->access,
+				 -1),
+			  -1),
+		   -1);
+
+      /* Remember this key. We don't want to ask again for key re-exchange */
+      spki_add_acl(self->db, acl);
       
       /* Write an ACL to disk. */
       if (self->file)
 	{
 	  A_WRITE(self->file, ssh_format("\n; ACL for host %lS\n", self->host->ip));
 	  A_WRITE(self->file,
-		  sexp_format(sexp_l(2, sexp_a(ATOM_ACL),
-				     sexp_l(3, sexp_a(ATOM_ENTRY),
-					    subject->key,
-					    sexp_l(2, sexp_a(ATOM_TAG),
-						   self->access,
-						   -1),
-					    -1),
-				     -1),
-			      SEXP_TRANSPORT, 0));
+		  sexp_format(acl, SEXP_TRANSPORT, 0));
 	  A_WRITE(self->file, ssh_format("\n"));
 	}
     }
@@ -481,75 +601,65 @@ make_lsh_host_db(struct spki_context *db,
   res->host = host;
   res->sloppy = sloppy;
   res->file = file;
-  res->hash = &sha1_algorithm;
+  res->hash = &crypto_sha1_algorithm;
 
   return &res->super;
 }
 
-/* Takes an spki_context as argument and returns a lookup_verifier */
-static void
-do_lsh_verifier(struct command *s,
-		struct lsh_object *a,
-		struct command_continuation *c,
-		struct exception_handler *e UNUSED)
+/* Takes options and an spki_context as argument and returns a
+ * lookup_verifier */
+
+DEFINE_COMMAND2(lsh_verifier_command)
+     (struct command_2 *s UNUSED,
+      struct lsh_object *a1,
+      struct lsh_object *a2,
+      struct command_continuation *c,
+      struct exception_handler *e UNUSED)
 {
-  CAST(options_command, self, s);
-  CAST_SUBTYPE(spki_context, db, a);
+  CAST(lsh_options, options, a1);
+  CAST_SUBTYPE(spki_context, db, a2);
   COMMAND_RETURN(c, make_lsh_host_db(db,
-				     self->options->super.tty,
-				     self->options->super.remote,
-				     self->options->sloppy,
-				     self->options->capture_file));
+				     options->super.tty,
+				     options->super.remote,
+				     options->sloppy,
+				     options->capture_file));
 }
 
-/* Takes an options object as argument and returns a lookup_verifier */
 
-DEFINE_COMMAND_SIMPLE(lsh_verifier_command, a)
+/* (login options public-keys connection) */
+DEFINE_COMMAND2(lsh_login_command)
+     (struct command_2 *s UNUSED,
+      struct lsh_object *a1,
+      struct lsh_object *a2,
+      struct command_continuation *c,
+      struct exception_handler *e UNUSED)
 {
-  CAST(lsh_options, options, a);
-
-  return
-    & make_options_command(options,
-			   do_lsh_verifier)->super;
-}
-
-/* list-of-public-keys -> login-command */
-static void
-do_lsh_login(struct command *s,
-	     struct lsh_object *a,
-	     struct command_continuation *c,
-	     struct exception_handler *e UNUSED)
-{
-  CAST(options_command, self, s);
-  CAST_SUBTYPE(object_list, keys, a);
+  CAST(lsh_options, options, a1);
+  CAST_SUBTYPE(object_list, keys, a2);
 
   struct client_userauth_method *password
-    = make_client_password_auth(self->options->super.tty);
+    = make_client_password_auth(options->super.tty);
+
+  /* FIXME: Perhaps we should try "none" only when using SRP. */
+  struct client_userauth_method *none
+    = make_client_none_auth();
   
   COMMAND_RETURN(c,
 		 make_client_userauth
-		 (ssh_format("%lz", self->options->super.user),
+		 (ssh_format("%lz", options->super.user),
 		  ATOM_SSH_CONNECTION,
 		  (LIST_LENGTH(keys)
 		   ? make_object_list
-		   (2, 
+		   (3,
+                    none,
 		    make_client_publickey_auth(keys),
 		    password,
 		    -1)
-		   : make_object_list(1, password, -1))));
+		   : make_object_list(2, none, password, -1))));
 }
 
-/* (login options public-keys connection) */
-DEFINE_COMMAND_SIMPLE(lsh_login_command, a)
-{
-  CAST(lsh_options, options, a);
 
-  return
-    &make_options_command(options,
-			  do_lsh_login)->super;
-}
-
-/* NOTE: options2identities is a command_simple, so it must not be
+/* NOTE: options2identities can block for reading, so it must not be
  * invoked directly. */
 
 /* Requests the ssh-userauth service, log in, and request connection
@@ -563,8 +673,8 @@ DEFINE_COMMAND_SIMPLE(lsh_login_command, a)
        (lambda (connection)
          (lsh_login options
 	   
-	   ; The prog1 delay is needed because options2identities is
-	   ; not a command_simple.
+	   ; The prog1 delay is needed because options2identities
+	   ; may not return immediately.
 	   (options2identities (prog1 options connection))
 
 	   ; Request the userauth service
@@ -574,9 +684,9 @@ DEFINE_COMMAND_SIMPLE(lsh_login_command, a)
    (expr
      (name make_lsh_connect)
      (params
-       (connect object command)
        (handshake object handshake_info)
-       (init object make_kexinit))
+       (init object make_kexinit)
+       (db object lookup_verifier))
      (expr (lambda (options)
                ; What to do with the service
 	       ((progn (options2actions options))
@@ -588,10 +698,9 @@ DEFINE_COMMAND_SIMPLE(lsh_login_command, a)
 		     ; Start the ssh transport protocol
 	             (connection_handshake
 		       handshake init
-		       (options2verifier options
-				         (options2known_hosts options))
+		       db
  		       ; Connect using tcp
-		       (connect (options2remote options)))))))))
+		       (connect_simple (options2remote options)))))))))
 */
 
 
@@ -614,7 +723,6 @@ const char *argp_program_bug_address = BUG_ADDRESS;
 
 #define OPT_DH 0x206
 #define OPT_SRP 0x207
-#define OPT_USERAUTH 0x208
 
 #define OPT_STDIN 0x210
 #define OPT_STDOUT 0x211
@@ -648,16 +756,20 @@ main_options[] =
     "Enable DH support (default, unless SRP is being used).", 0 },
 
   { "no-dh-keyexchange", OPT_DH | ARG_NOT, NULL, 0, "Disable DH support.", 0 },
-
-  { "userauth", OPT_USERAUTH, NULL, 0,
-    "Request the ssh-userauth service (default, unless SRP is being used).", 0 },
-  { "no-userauth", OPT_USERAUTH | ARG_NOT, NULL, 0,
-    "Request the ssh-userauth service (default if SRP is used).", 0 },
-
-  /* ACtions */
+  
+  /* Actions */
   { "forward-remote-port", 'R', "remote-port:target-host:target-port",
     0, "", CLIENT_ARGP_ACTION_GROUP },
   { "gateway", 'G', NULL, 0, "Setup a local gateway", 0 },
+
+  /* FIXME: Perhaps this should be moved from lsh.c to client.c? It
+   * doesn't work with lshg. Or perhaps that can be fixed?
+   * About the same problem applies to -R. */
+  
+  { "x11-forward", 'x', NULL, 0, "Enable X11 forwarding.", CLIENT_ARGP_MODIFIER_GROUP },
+  { "no-x11-forward", 'x' | ARG_NOT, NULL, 0,
+    "Disable X11 forwarding (default).", 0 },
+  
   { NULL, 0, NULL, 0, NULL, 0 }
 };
 
@@ -721,6 +833,9 @@ main_argp_parser(int key, char *arg, struct argp_state *state)
 	  break;
 	}
 
+      if (!self->super.random)
+	argp_failure( state, EXIT_FAILURE, 0,  "No randomness generator available.");
+
       if (self->with_dh_keyexchange < 0)
 	self->with_dh_keyexchange = !self->with_srp_keyexchange;
       
@@ -736,7 +851,7 @@ main_argp_parser(int key, char *arg, struct argp_state *state)
 	      LIST(self->kex_algorithms)[i++] = ATOM_SRP_RING1_SHA1_LOCAL;
 	      ALIST_SET(self->algorithms->algorithms,
 			ATOM_SRP_RING1_SHA1_LOCAL,
-			&make_srp_client(make_srp1(&self->random->super),
+			&make_srp_client(make_srp1(self->super.random),
 					 self->super.tty,
 					 ssh_format("%lz", self->super.user))
 			->super);
@@ -747,7 +862,7 @@ main_argp_parser(int key, char *arg, struct argp_state *state)
 	      LIST(self->kex_algorithms)[i++] = ATOM_DIFFIE_HELLMAN_GROUP1_SHA1;
 	      ALIST_SET(self->algorithms->algorithms,
 			ATOM_DIFFIE_HELLMAN_GROUP1_SHA1,
-			&make_dh_client(make_dh1(&self->random->super))
+			&make_dh_client(make_dh1(self->super.random))
 			->super);
 	    }
 	}
@@ -755,49 +870,8 @@ main_argp_parser(int key, char *arg, struct argp_state *state)
 	argp_error(state, "All keyexchange algorithms disabled.");
 
       {
-	struct command *perform_userauth = NULL;
-
-	/* NOTE: The default, self->with_userauth < 0, means that we
-	 * should figure out the right thing automatically. */
-
-	if (self->with_userauth < 0)
-	  {
-	    /* Make the decision  early, if possible. */
-	    if (!self->with_srp_keyexchange)
-	      /* We're not using SRP, so we request the
-	       * ssh-userauth-service */
-	      self->with_userauth = 1;
-
-	    else if (!self->with_dh_keyexchange)
-	      /* We're using SRP, and should not need any extra
-	       * user authentication. */
-	      self->with_userauth = 0;
-	  }
-	   
-	if (self->with_userauth)
-	  {
-	    CAST_SUBTYPE(command, o, make_lsh_userauth(self));
-	    perform_userauth = o;
-	  }
-
-	switch(self->with_userauth)
-	  {
-	  case 0:
-	    self->service = &request_connection_service.super;
-	    break;
-	  case 1:
-	    self->service = perform_userauth;
-	    break;
-	  case -1:
-	    /* Examine the CONNECTION_SRP flag, later. */
-	    self->service
-	      = make_connection_if_srp(&request_connection_service.super,
-				       perform_userauth);
-	  default:
-	    fatal("Internal error.\n");
-	  }
-	  
-	assert(self->service);
+        CAST_SUBTYPE(command, o, make_lsh_userauth(self));
+        self->service = o;
       }
 	
       {
@@ -808,18 +882,21 @@ main_argp_parser(int key, char *arg, struct argp_state *state)
 	  s = self->capture;
 	else if (self->sloppy)
 	  {
-	    tmp = ssh_cformat("%lz/.lsh/captured_keys", self->home);
-	    s = tmp->data;
+	    tmp = ssh_format("%lz/.lsh/captured_keys", self->home);
+	    s = lsh_get_cstring(tmp);
 	  }
 	if (s)
 	  {
+	    static const struct report_exception_info report =
+	      STATIC_REPORT_EXCEPTION_INFO(EXC_IO, EXC_IO,
+					   "Writing new ACL: ");
+	    
 	    struct lsh_fd *f
-	      = io_write_file(self->super.backend, s,
+	      = io_write_file(s,
 			      O_CREAT | O_APPEND | O_WRONLY,
 			      0600, 500, NULL,
 			      make_report_exception_handler
-			      (make_report_exception_info(EXC_IO, EXC_IO,
-							  "Writing new ACL: "),
+			      (&report,
 			       &default_exception_handler,
 			       HANDLER_CONTEXT));
 	    if (f)
@@ -858,7 +935,7 @@ main_argp_parser(int key, char *arg, struct argp_state *state)
 	      
 	  client_add_action(&self->super,
 			    make_gateway_setup
-			    (make_listen_local(self->super.backend, gateway)));
+			    (make_listen_local(gateway)));
 	}
 
       if (object_queue_is_empty(&self->super.actions))
@@ -866,10 +943,7 @@ main_argp_parser(int key, char *arg, struct argp_state *state)
 	  argp_error(state, "No actions given.");
 	  break;
 	}
-      
-      /* Start background poll */
-      RANDOM_POLL_BACKGROUND(self->random->poller);
-      
+
       break;
       
     case 'i':
@@ -897,28 +971,6 @@ main_argp_parser(int key, char *arg, struct argp_state *state)
     CASE_FLAG(OPT_DH, with_dh_keyexchange);
     CASE_FLAG(OPT_SRP, with_srp_keyexchange);
 
-    CASE_FLAG(OPT_USERAUTH, with_userauth);
-
-#if 0
-    case 'L':
-      {
-	UINT32 listen_port;
-	struct address_info *target;
-
-	if (!client_parse_forward_arg(arg, &listen_port, &target))
-	  argp_error(state, "Invalid forward specification '%s'.", arg);
-
-	client_add_action(&self->super, make_forward_local_port
-			  (self->super.backend,
-			   make_address_info((self->super.with_remote_peers
-					      ? NULL
-					      : ssh_format("%lz", "127.0.0.1")),
-					     listen_port),
-			   target));
-	break;
-      }      
-#endif
-      
     case 'R':
       {
 	UINT32 listen_port;
@@ -928,8 +980,7 @@ main_argp_parser(int key, char *arg, struct argp_state *state)
 	  argp_error(state, "Invalid forward specification '%s'.", arg);
 
 	client_add_action(&self->super, make_forward_remote_port
-			  (self->super.backend,
-			   make_address_info((self->super.with_remote_peers
+			  (make_address_info((self->super.with_remote_peers
 					      ? ssh_format("%lz", "0.0.0.0")
 					      : ssh_format("%lz", "127.0.0.1")),
 					     listen_port),
@@ -940,6 +991,7 @@ main_argp_parser(int key, char *arg, struct argp_state *state)
       }      
       
     CASE_FLAG('G', start_gateway);
+    CASE_FLAG('x', super.with_x11);
     }
 
   return 0;
@@ -978,7 +1030,7 @@ do_lsh_default_handler(struct exception_handler *s,
       CAST_SUBTYPE(io_exception, exc, e);
       *self->status = EXIT_FAILURE;
       
-      werror("lsh: %z, (errno = %i)\n", e->msg, exc->error);
+      werror("%z, (errno = %i)\n", e->msg, exc->error);
     }
   else
     switch(e->type)
@@ -991,7 +1043,7 @@ do_lsh_default_handler(struct exception_handler *s,
       case EXC_CHANNEL_REQUEST:
       case EXC_CHANNEL_OPEN:
 
-	werror("lsh: %z\n", e->msg);
+	werror("%z\n", e->msg);
 	*self->status = EXIT_FAILURE;
 	break;
       default:
@@ -1018,7 +1070,8 @@ make_lsh_default_handler(int *status, struct exception_handler *parent,
 int main(int argc, char **argv)
 {
   struct lsh_options *options;
-
+  struct spki_context *spki;
+  
   /* Default exit code if something goes wrong. */
   int lsh_exit_code = 17;
 
@@ -1026,7 +1079,7 @@ int main(int argc, char **argv)
     = make_lsh_default_handler(&lsh_exit_code, &default_exception_handler,
 			       HANDLER_CONTEXT);
 
-  struct io_backend *backend = make_io_backend();
+  io_init();
   
   /* For filtering messages. Could perhaps also be used when converting
    * strings to and from UTF8. */
@@ -1035,27 +1088,42 @@ int main(int argc, char **argv)
   /* FIXME: Choose character set depending on the locale */
   set_local_charset(CHARSET_LATIN1);
 
-  options = make_options(backend, handler, &lsh_exit_code);
+  options = make_options(handler, &lsh_exit_code);
 
+  if (!options)
+    return EXIT_FAILURE;
+  
   argp_parse(&main_argp, argc, argv, ARGP_IN_ORDER, NULL, options);
+
+  if (!options->super.random)
+    {
+      werror("Failed to initialize randomness generator.\n");
+      return EXIT_FAILURE;
+    }
+
+  spki = read_known_hosts(options);
   
   {
     struct lsh_object *o =
       make_lsh_connect(
-	make_simple_connect(backend, NULL),
 	make_handshake_info(CONNECTION_CLIENT,
 			    "lsh - a free ssh", NULL,
 			    SSH_MAX_PACKET,
-			    &options->random->super,
+			    options->super.random,
 			    options->algorithms->algorithms,
 			    NULL),
-	make_simple_kexinit(&options->random->super,
+	make_simple_kexinit(options->super.random,
 			    options->kex_algorithms,
 			    options->algorithms->hostkey_algorithms,
 			    options->algorithms->crypto_algorithms,
 			    options->algorithms->mac_algorithms,
 			    options->algorithms->compression_algorithms,
-			    make_int_list(0, -1)));
+			    make_int_list(0, -1)),
+        make_lsh_host_db(spki,
+                         options->super.tty,
+                         options->super.remote,
+                         options->sloppy,
+                         options->capture_file));
     
     CAST_SUBTYPE(command, lsh_connect, o);
 
@@ -1070,15 +1138,15 @@ int main(int argc, char **argv)
    * or not anybody is still using it. */
   close(STDOUT_FILENO);
   if (open("/dev/null", O_WRONLY) != STDOUT_FILENO)
-    werror("lsh: Strange: Final redirect of stdout to /dev/null failed.\n");
+    werror("Strange: Final redirect of stdout to /dev/null failed.\n");
 #endif
   
-  io_run(backend);
+  io_run();
+
+  /* Close all files and other resources associated with the backend. */
+  io_final();
   
   /* FIXME: Perhaps we have to reset the stdio file descriptors to
    * blocking mode? */
-
-  gc_final();
-  
   return lsh_exit_code;
 }
